#!/usr/bin/env python

# Prepare environment
import sys
from pika.frame import Body
pyxml=None
index = 0
for p in sys.path:
    if -1 != p.find('pyxml'):
         pyxml = p
    index += 1

if None != pyxml:
    sys.path.remove(pyxml)

import errno
def mkdirs(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise e

FILENAME_LENGTH_LIMIT = 255

import hashlib
import json
import os
import socket
import subprocess
import xml.etree.ElementTree as ET

import pp_engine

from rabbitmq_consume import StopWorkerException, StopListeningException

class PSquaredConsumption(object):

    def __init__(self,
                 properties,
                 body,
                 redelivered):
        self._body = body
        self._properties = properties
        # If redelivery is set the following to `redelivered` less set to `False`
        self._redelivered = False # redelivered

    def executionMessage(self):
        return 'Executing on ' + socket.gethostname()

    def consume(self):
        try:
            document = ET.fromstring(self._body)
            action = document.get('action')
            if 'stop_listening' == action:
                raise StopListeningException()
            if 'stop_worker' == action:
                raise StopWorkerException()
            if 'ignore' == action:
                print ('Ignoring execution request')
            else:
                element = document.find('submission_uri')
                if None != element:
                    submission_uri = element.text
                    cmd = document.find('command').text
                    args = document.findall('argument')
                    if 0 == len(args):
                        arguments = None
                    else:
                        arguments = []
                        for arg in args:
                            arguments.append(arg.text)
                    element = document.find('success_cmd')
                    if None == element or '' == element.text:
                        success_cmd = None
                    else:
                        success_cmd = element.text
                    element = document.find('failure_cmd')
                    if None == element or '' == element.text:
                        failure_cmd = None
                    else:
                        failure_cmd = element.text
                    attribute = document.get('stdout')
                    if None == attribute or '' == attribute:
                        cmd_out = None
                    else:
                        if attribute.startswith('/'):
                            filepath = attribute
                        else:
                            filepath = os.path.expanduser('~/' + attribute)
                        outDir = os.path.dirname(filepath)
                        if None != outDir and '' != outDir:
                            mkdirs(outDir)

                        # Temporary support for file name limit of FILENAME_LENGTH_LIMIT
                        outBase = os.path.basename(filepath)
                        if FILENAME_LENGTH_LIMIT < len(outBase):
                            outBase=str(hashlib.sha256(outBase).hexdigest())
                            if None != outDir and '' != outDir:
                                filepath=os.path.join(outDir, outBase)
                            else:
                                filepath=outBase
                        # End temporary support

                        cmd_out = open(filepath, 'w')
                    attribute = document.get('stderr')
                    if None == attribute or '' == attribute:
                        cmd_err = None
                    else:
                        if attribute.startswith('/'):
                            filepath = attribute
                        else:
                            filepath = os.path.expanduser('~/' + attribute)
                        errDir = os.path.dirname(filepath)
                        if None != errDir and '' != errDir:
                            mkdirs(errDir)

                        # Temporary support for file name limit of FILENAME_LENGTH_LIMIT
                        errBase = os.path.basename(filepath)
                        if FILENAME_LENGTH_LIMIT < len(errBase):
                            errBase=str(hashlib.sha256(errBase).hexdigest())
                            if None != errDir and '' != errDir:
                                filepath=os.path.join(errDir, errBase)
                            else:
                                filepath=errBase
                        # End temporary support

                        cmd_err = open(filepath, 'w')
                    print ('Beginning state machine completion')
                    pp_engine.execute(self.executionMessage(),
                                      cmd,
                                      arguments,
                                      submission_uri,
                                      cmd_out,
                                      cmd_err,
                                      success_cmd,
                                      failure_cmd,
                                      self._redelivered)
                    print ('Completed state machine')
        except StopListeningException as e:
            raise e
        except StopWorkerException as e:
            raise e
        except pp_engine.PSquaredFailure as e:
            print ('Failed to complete state machine')
            print ('Failed in PSquared:\n' + e.message, file=sys.stderr)
            #print >> sys.stderr, 'Failed in PSquared:\n' + e.message
            print ('END CONSUMPTION')
        except Exception as e:
            print ('Failed to complete state machine')
            print (e, file=sys.stderr)
            #print (e.message, file=sys.stderr)
            #print >> sys.stderr, e.message
            print ('ABORT CONSUMPTION')



#!/usr/bin/env python

from __future__ import print_function

# Prepare environment
import sys
pyxml=None
index = 0
for p in sys.path:
    if -1 != p.find('pyxml'):
         pyxml = p
    index += 1

if None != pyxml:
    sys.path.remove(pyxml)

import xml.etree.ElementTree as ET

import os
cert_dir = os.getenv('HOME') + '/.psquared/client/certs'

import requests
session=requests.Session()
adapter = requests.adapters.HTTPAdapter(pool_connections=32, pool_maxsize=32)
session.mount('http://', adapter)

CERT = cert_dir + '/psquare_client.crt' #Client certificate
KEY = cert_dir + '/psquare_client.key' #Client private key
CACERT = cert_dir + '/psquare_cacert.crt' #CA certificates

if os.path.exists(CERT) and os.path.exists(CERT):
    session.cert = (CERT, KEY)
if os.path.exists(CACERT):
    session.verify = CACERT

HEADERS = {'Content-Type': 'application/xml',
           'Accept': 'application/xml'}
DEBUG = False
RESUBMISSION_LIMIT = 8

def eprint(*args, **kwargs):
    """Prints to standard error"""
    print(*args, file=sys.stderr, **kwargs)


# Simple pretty print for *IX
def pretty_print(uri, s, response = True):
    if DEBUG:
        if None != uri:
            if response:
                eprint('URI : Response : ' + uri)
            else:
                eprint('URI : Request :  ' + uri)
        os.system("echo '" + str(s) + "' | xmllint -format - 1>&2")
        eprint('--------')


class FatalError(Exception):
    def __init__(self, message, errorCode, response):
        self.code = errorCode
        self.message = message
        self.response = response


def checkStatus(uri, r, expected):
    if 400 == r.status_code:
        raise FatalError('Application at "' + uri  + '" requires SSL certificate', r.status_code, r.text)
    elif 401 == r.status_code:
        raise FatalError('Not authorized to execute commands for Application at "' + uri, r.status_code, r.text)
    elif 404 == r.status_code:
        raise FatalError('Application at "' + uri  + '" not found', r.status_code, r.text)
    elif expected != r.status_code:
        raise FatalError('Unexpected status (' + str(r.status_code) + ') returned from "' + uri  + '"', r.status_code, r.text)
    return


class PSquaredFailure(Exception):
    """
    Thrown when a RESTful does not complete as expected.
    """
    def __init__(self, message, errorCode):
        # Call the base class constructor with the parameters it needs
        Exception.__init__(self, message)
        self.code = errorCode
        self.message = message # JM 08/25/2020 


def get_attachment(message):
    """
    Returns the 'attachment' document that includes the supplied message, if any.
    """
    attachment = ET.Element('attachment')
    if None != message:
        msg = ET.Element('message')
        msg.text = message
        attachment.append(msg)
    return attachment


def execute_transition(uri, message):
    """
    Requests the PSquared instance create a new realized state.
    """
    # Prepare attachment document
    attachment = get_attachment(message)
    pretty_print(uri, ET.tostring(attachment), False)
    r = session.post(uri, data=ET.tostring(attachment), headers=HEADERS)
    if 404 == r.status_code:
        raise PSquaredFailure('Exit request "' + uri  + '" not found', r.status_code)
    elif 409 == r.status_code:
        raise PSquaredFailure('Another process has changed the processing of this request, so this attempt will halt', r.status_code)
    elif 201 != r.status_code:
        raise PSquaredFailure('Unexpected status (' + str(r.status_code) + ') returned from "' + uri  + '"', r.status_code)
    pretty_print(uri, r.text)
    return ET.fromstring(r.text)


def parse_submitted_response(state, resubmission_count):
    e=state.findall('exits/exit')
    if 0 == len(e):
        e=state.find('exited')
        if None == e:
            raise PSquaredFailure('Incomplete response returned from "' + uri  + '"', 1)
        if not 0 > resubmission_count:
            if 'executing' == e.find('name').text:
                if resubmission_count == RESUBMISSION_LIMIT:
                    raise PSquaredFailure('re-submission limit has been reached, so this attempt will halt', 409)
                return automatic_resubmission(e.find('uri').text, resubmission_count + 1)
            raise PSquaredFailure('State that follows submission does not allow automatic re-submission, so this attempt will halt', 409)
        raise PSquaredFailure('Another process has started processing this request, so this attempt will halt', 409)
    executingUri = None
    for exit in e:
        if 'executing' == exit.find('name').text:
            return exit.find('uri').text
    raise PSquaredFailure('Unexpected response returned from "' + uri  + '"', 1)


def automatic_resubmission(uri, resubmission_count):
    """
    Resubmits this request as its previous engine failed top complete.
    """
    try:
        r = session.get(uri)
    except ConnectionError as e:
        raise PSquaredFailure('Could not connect to server', -1)
    checkStatus(uri, r, 200)
    pretty_print(uri, r.text)
    state = ET.fromstring(r.text)
    e=state.findall('exits/exit')
    if 0 == len(e):
        e=state.find('exited')
        if None == e:
            raise PSquaredFailure('Incomplete response returned from "' + uri  + '"', 1)
        if 'reschedule' == e.find('name').text:
            return get_execution_uri(e.find('uri').text, resubmission_count)
        raise PSquaredFailure('Another process has processed this request, so this attempt will halt', 409)
    for exit in e:
       if 'reschedule' == exit.find('name').text:
           rescheduleUri = exit.find('uri').text
           break

    state = execute_transition(rescheduleUri,
                               'Rescheduling as previous execution died')
    e=state.findall('exits/exit')
    if 0 == len(e):
        raise PSquaredFailure('Failed to correctly reschedule submission')
    return parse_submitted_response(state, resubmission_count)


def get_execution_uri(uri, resubmission_count):
    """
    Returns the URI that will signal to PSqaured that processing has begun.
    """
    try:
        r = session.get(uri)
    except ConnectionError as e:
        raise PSquaredFailure('Could not connect to server', -1)
    checkStatus(uri, r, 200)
    pretty_print(uri, r.text)
    state = ET.fromstring(r.text)
    return parse_submitted_response(state,
                                    resubmission_count)
    

def processing_beginning(uri,
                         message):
    """
    Informs the PSquared instance that processing is beginning for this
    item/version pairing.
    """
    state = execute_transition(uri,
                               message)
    e=state.findall('exits/exit')
    for exit in e:
        if 'processed' == exit.find('name').text:
            processedUri = exit.find('uri').text
            break
    
    for exit in e:
        if 'failed' == exit.find('name').text:
            failedUri = exit.find('uri').text
            break
    
    return processedUri, failedUri


def process(cmd,
            params,
            cmd_out,
            cmd_err):
    """
    Processes the supplied command, return any message created by the command while
    it was executing.
    """
    import subprocess
    import tempfile
    args = []
    args.append(cmd)
    args.extend(params)
    fd, message_file = tempfile.mkstemp(text=True)
    cmd_env = os.environ.copy()
    cmd_env['PP_MESSAGE_FILE'] = message_file
    proc = subprocess.Popen(args,
                            env=cmd_env,
                            stdout=cmd_out,
                            stderr=cmd_err)
    result = proc.wait()
    message = os.fdopen(fd, 'r').read()
    os.remove(message_file)
    return result, message.strip()


def execute(message_for_execute,
            cmd,
            arguments,
            submission_uri,
            cmd_out,
            cmd_err,
            success_cmd,
            failure_cmd,
            resubmission):
    if resubmission:
        resubmission_count = 0
    else:
        resubmission_count = -1
    executingUri = get_execution_uri(submission_uri,
                                     resubmission_count)
    processedUri, failedUri = processing_beginning(executingUri,
                                                   message_for_execute)
    if None == arguments:
        params = []
    else:
        params = arguments
    if None == cmd:
        result, message = 0, 'No command supplied'
    else:
        try:
            result, message = process(cmd,
                                      params,
                                      cmd_out,
                                      cmd_err)
        except OSError as o:
            eprint(cmd, params)
            eprint('Failed to execute command')
            if None != o.errno:
                eprint('Error Number: ' + str(o.errno))
            if None != o.strerror:
                eprint('Explanation: ' + o.strerror)
            if None != o.filename:
                eprint('Filename: ' + o.filename)
            result, message = 255, 'Failed to execute command, check log file for details'
    if 0 == result:
        execute_transition(processedUri,
                           message)
        if None != success_cmd:
            try:
                process(success_cmd,
                        params,
                        cmd_out,
                        cmd_err)
            except OSError as o:
                raise PSquaredFailure('Failed to execute "success" command', 255)
    else:
        execute_transition(failedUri,
                           message)
        if None != failure_cmd:
            try:
                process(failure_cmd,
                        params,
                        cmd_out,
                        cmd_err)
            except OSError as o:
                raise PSquaredFailure('Failed to execute "failure" command', 255)


if __name__ == '__main__':
    import argparse
    import sys
    import socket
    import logging

    parser = argparse.ArgumentParser(description='Executes a command with the PSquared State machine.')
    parser.add_argument('-d',
                      '--debug',
                      dest='DEBUG',
                      help='print out detail information to stdout.',
                      action='store_true',
                      default=False)
    parser.add_argument('-e',
                        dest='ERROR',
                        help='The file, as opposed to stderr, into which to write errors from the command')
    parser.add_argument('--log_file',
                        dest='LOG_FILE',
                        help='The file, as opposed to stdout, into which to write log messages from the state machine')
    parser.add_argument('-o',
                        dest='OUTPUT',
                        help='The file, as opposed to stdout, into which to write output from the command')
    parser.add_argument('-s',
                      '--success_cmd',
                      dest = 'SUCCESS_CMD',
                      help = 'The command to run after a successful processing',
                      default = None)
    parser.add_argument('-f',
                      '--failure_cmd',
                      dest = 'FAILURE_CMD',
                      help = 'The command to run after a failed processing',
                      default = None)
    parser.add_argument('-r',
                      '--resubmission',
                      dest='RESUBMISSION',
                      help='allows automatic re-submission to take place.',
                      action='store_true',
                      default=False)
    parser.add_argument('submission_url',
                        help='The submission URL used to create the processing request to be handled')
    parser.add_argument('cmd',
                        help='The command to be executed')
    parser.add_argument('args',
                        nargs=argparse.REMAINDER,
                        help='An argument to the command to be executed')
    options = parser.parse_args()

    if options.DEBUG:
        log_level = logging.INFO
        DEBUG = True
    else:
        log_level = logging.WARNING
    if options.LOG_FILE:
        logging.basicConfig(filename=options.LOG_FILE,
                            level=log_level)
    else:
        logging.basicConfig(Stream=sys.stdout,
                            level=log_level)

    print('Beginning state machine execution')
    try:
        execute('Executing on ' + socket.gethostname() + ' using pp_engine.py',
                options.cmd,
                options.args,
                options.submission_url,
                options.OUTPUT,
                options.ERROR,
                options.SUCCESS_CMD,
                options.FAILURE_CMD,
                options.RESUBMISSION)
        print('Completed state machine execution')
    except FatalError as e:
        eprint(e.message)
        sys.exit(e.code)



"""A setuptools based setup module.

See:
https://gitlab.com/nest.lbl.gov/psquared
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path
# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='psquared_runner',  # Required
    version='2.3.1',  # Required
    description='Manage the connection and communications to a PSquared server.',  # Optional
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    url='http://nest.lbl.gov/projects/psquared',  # Optional
    author='Simon Patton',  # Optional
    author_email='sjpatton@lbl.gov',  # Optional
    classifiers=[  # Optional
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    #keywords='keyword1 keyword2',  # Optional
    packages=['psquared_runner'],  # Required
    py_modules=['pp_consume'],  # Optional
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    install_requires=['requests'],  # Optional
    entry_points={  # Optional
        'console_scripts': [
            'pp-cli=pp_cli:main',
        ],
    }
)
